##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""WSGI tests"""
import doctest
import io
import re
import unittest

import zope.component
import zope.component.testing
import zope.event
import zope.interface
from zope.authentication.interfaces import IUnauthenticatedPrincipal
from zope.component.testlayer import ZCMLFileLayer
from zope.publisher.interfaces.logginginfo import ILoggingInfo
from zope.testing import renormalizing

import zope.app.wsgi
from zope.app.wsgi.testing import SillyMiddleWare
from zope.app.wsgi.testlayer import BrowserLayer


def creating_app_w_paste_emits_ProcessStarting_event():
    """
    >>> import zope.event
    >>> events = []
    >>> subscriber = events.append
    >>> zope.event.subscribers.append(subscriber)

    >>> import os, tempfile
    >>> temp_dir = tempfile.mkdtemp()
    >>> sitezcml = os.path.join(temp_dir, 'site.zcml')
    >>> with open(sitezcml, 'w') as f:
    ...     _ = f.write('<configure />')
    >>> zopeconf = os.path.join(temp_dir, 'zope.conf')
    >>> with open(zopeconf, 'w') as f:
    ...     _ = f.write('''
    ... site-definition %s
    ...
    ... <zodb>
    ...   <mappingstorage />
    ... </zodb>
    ...
    ... <eventlog>
    ...   <logfile>
    ...     path STDOUT
    ...   </logfile>
    ... </eventlog>
    ... ''' % sitezcml)

    >>> import zope.app.wsgi.paste, zope.processlifetime
    >>> app = zope.app.wsgi.paste.ZopeApplication(
    ...     {}, zopeconf, handle_errors=False)

    >>> len([e for e in events
    ...     if isinstance(e, zope.processlifetime.ProcessStarting)]) == 1
    True

    >>> zope.event.subscribers.remove(subscriber)
    """


wsgiapp_layer = BrowserLayer(zope.app.wsgi, name='wsgiapp', allowTearDown=True)


def setUpWSGIApp(test):
    test.globs['wsgi_app'] = wsgiapp_layer.make_wsgi_app()


def setUpSillyWSGIApp(test):
    test.globs['wsgi_app'] = wsgiapp_layer.make_wsgi_app(SillyMiddleWare)


@zope.component.adapter(IUnauthenticatedPrincipal)
@zope.interface.implementer(ILoggingInfo)
def could_not_adapt_principal_to_logging_info(context):
    """Fake that a principal could not be adapted to ILoggingInfo."""
    return None


class WSGIPublisherApplicationTests(unittest.TestCase):
    """Testing .WSGIPublisherApplication."""

    layer = wsgiapp_layer

    def setUp(self):
        zope.component.provideAdapter(
            could_not_adapt_principal_to_logging_info)

    def tearDown(self):
        super().tearDown()
        gsm = zope.component.getGlobalSiteManager()
        assert gsm.unregisterAdapter(could_not_adapt_principal_to_logging_info)

    def test_WSGIPublisherApplication___call___1(self):
        """It sets '-' as 'wsgi.logging_info' in environ as fall back.

        This is the case if the principal couldn't be adapted to ILoggingInfo.
        """
        from . import WSGIPublisherApplication

        app = WSGIPublisherApplication()
        environ = {'wsgi.input': io.BytesIO(b'')}
        list(app(environ, lambda status, headers: None))
        self.assertEqual('-', environ['wsgi.logging_info'])


class AuthHeaderTestCase(unittest.TestCase):

    def test_auth_encoded(self):
        from zope.app.wsgi.testlayer import auth_header
        header = 'Basic Z2xvYmFsbWdyOmdsb2JhbG1ncnB3'
        self.assertEqual(auth_header(header), header)

    def test_auth_non_encoded(self):
        from zope.app.wsgi.testlayer import auth_header
        header = 'Basic globalmgr:globalmgrpw'
        expected = 'Basic Z2xvYmFsbWdyOmdsb2JhbG1ncnB3'
        self.assertEqual(auth_header(header), expected)

    def test_auth_non_encoded_empty(self):
        from zope.app.wsgi.testlayer import auth_header
        header = 'Basic globalmgr:'
        expected = 'Basic Z2xvYmFsbWdyOg=='
        self.assertEqual(auth_header(header), expected)
        header = 'Basic :pass'
        expected = 'Basic OnBhc3M='
        self.assertEqual(auth_header(header), expected)

    def test_auth_non_encoded_colon(self):
        from zope.app.wsgi.testlayer import auth_header
        header = 'Basic globalmgr:pass:pass'
        expected = 'Basic Z2xvYmFsbWdyOnBhc3M6cGFzcw=='
        self.assertEqual(auth_header(header), expected)


class TestFakeResponse(unittest.TestCase):

    def test_doesnt_assume_encoding_of_headers(self):
        # https://github.com/zopefoundation/zope.app.wsgi/issues/7
        # For BWC, we allow them to be bytes or str.
        # The body/__str__ can be decoded correctly too when this happens
        from zope.app.wsgi.testlayer import FakeResponse

        class MockResponse:

            status = '200 OK'
            body = ''

            def __init__(self):
                self.headerlist = []

        response = MockResponse()
        # A latin-1 byte.
        response.headerlist.append((b"X-Header",
                                    b"voill\xe0"))

        fake = FakeResponse(response)
        self.assertEqual(
            fake.getOutput(),
            b'HTTP/1.0 200 OK\nX-Header: voill\xe0')
        # No matter the platform, str/bytes should not raise
        self.assertIn('HTTP', str(fake))
        self.assertIn(b'HTTP', bytes(fake))
        self.assertEqual(str(fake),
                         'HTTP/1.0 200 OK\nX-Header: voill\xe0')

        # A utf-8 byte, smuggled inside latin-1, as discussed in PEP3333
        response.headerlist[0] = (
            b'X-Header',
            'p-o-p \U0001F4A9'.encode().decode('latin-1'))
        self.assertEqual(fake.getOutput(),
                         b'HTTP/1.0 200 OK\nX-Header: p-o-p \xf0\x9f\x92\xa9')
        self.assertIn('HTTP', str(fake))
        self.assertIn(b'HTTP', bytes(fake))
        self.assertEqual(str(fake),
                         'HTTP/1.0 200 OK\nX-Header: p-o-p \xf0\x9f\x92\xa9')


def test_suite():
    suites = []
    checker = renormalizing.RENormalizing([
        (re.compile(
            r"&lt;class 'zope.component.interfaces.ComponentLookupError'&gt;"),
         r'ComponentLookupError'),
    ])

    filereturns_suite = doctest.DocFileSuite(
        'filereturns.txt', setUp=setUpWSGIApp)
    filereturns_suite.layer = wsgiapp_layer
    suites.append(filereturns_suite)

    dt_suite = doctest.DocTestSuite()
    dt_suite.layer = wsgiapp_layer
    suites.append(dt_suite)

    readme_test = doctest.DocFileSuite(
        'README.txt',
        checker=checker,
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)
    # This test needs its own layer/teardown, since it registers components
    # with objects that later do not exist.
    readme_test.layer = ZCMLFileLayer(zope.app.wsgi, name="README")
    suites.append(readme_test)

    doctest_suite = doctest.DocFileSuite(
        'fileresult.txt', 'paste.txt',
        checker=checker,
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)
    doctest_suite.layer = ZCMLFileLayer(zope.app.wsgi)
    suites.append(doctest_suite)

    testlayer_suite = doctest.DocFileSuite(
        'testlayer.txt', setUp=setUpSillyWSGIApp,
        optionflags=doctest.NORMALIZE_WHITESPACE)
    testlayer_suite.layer = wsgiapp_layer
    suites.append(testlayer_suite)

    suites.append(unittest.defaultTestLoader.loadTestsFromName(__name__))

    return unittest.TestSuite(suites)



##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
# This package is developed by the Zope Toolkit project, documented here:
# http://docs.zope.org/zopetoolkit
# When developing and releasing this package, please follow the documented
# Zope Toolkit policies as described by this documentation.
##############################################################################
"""Setup for zope.app.wsgi package
"""
import os

from setuptools import find_packages
from setuptools import setup


TESTS_REQUIRE = [
    'WebTest',
    'zope.authentication',
    'zope.browserpage',
    'zope.principalregistry >=4.0.0a1',
    'zope.securitypolicy >=4.0.0a1',
    'zope.testing',
    'zope.testrunner',
]


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        return f.read()


setup(
    name='zope.app.wsgi',
    version='5.3',
    url='https://github.com/zopefoundation/zope.app.wsgi',
    project_urls={
        'Issue Tracker': ('https://github.com/zopefoundation/'
                          'zope.app.wsgi/issues'),
        'Sources': 'https://github.com/zopefoundation/zope.app.wsgi',
    },
    license='ZPL 2.1',
    description='WSGI application for the zope.publisher',
    long_description=read('README.rst') +
            '\n\n' +
            read('CHANGES.rst'),
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.org',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Framework :: Zope :: 3',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['zope', 'zope.app'],
    extras_require=dict(
        test=TESTS_REQUIRE,
        testlayer=['WebTest'],
    ),
    python_requires='>=3.8',
    install_requires=[
        'setuptools',
        'ZConfig',
        'legacy-cgi; python_version >= "3.13"',  # WebOb uses the cgi module
        'transaction',
        'zope.app.appsetup >= 4.0',
        'zope.processlifetime',
        'zope.app.publication >= 4.0',
        'zope.event',
        'zope.interface',
        'zope.publisher>=4.0.0a3',
        'zope.security>4.0.0a2',
        'zope.component',
        'zope.container >=4.0.0a1',
        'zope.site >= 4.0.0a1',
        'zope.traversing>=4.0.0a1',
    ],
    tests_require=TESTS_REQUIRE,
    entry_points={
        'paste.app_factory': [
            'main = zope.app.wsgi.paste:ZopeApplication'
        ]
    },
    include_package_data=True,
    zip_safe=False,
)



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import base64
import http.client as httpclient
import re
import typing
import xmlrpc.client
from io import BytesIO

import transaction
import webtest
from webtest import TestRequest
from zope.app.appsetup.testlayer import ZODBLayer

from zope.app.wsgi import WSGIPublisherApplication


basicre = re.compile('Basic (.+)?:(.+)?$')
_TEST_APP_FOR_ENCODING = webtest.TestApp(None)


def auth_header(header):
    """This function takes an authorization HTTP header and encode the
    couple user, password into base 64 like the HTTP protocol wants
    it.
    """
    match = basicre.match(header)
    if match:
        u, p = match.group(1, 2)
        if u is None:
            u = ''
        if p is None:
            p = ''
        plain = f'{u}:{p}'
        auth = base64.b64encode(plain.encode('utf-8'))
        return 'Basic %s' % str(auth.rstrip().decode('latin1'))
    return header


def is_wanted_header(header):
    """Return True if the given HTTP header key is wanted.
    """
    key, value = header
    return key.lower() not in ('x-content-type-warning', 'x-powered-by')


class TransactionMiddleware:
    """This middleware makes the WSGI application compatible with the
    HTTPCaller behavior defined in zope.app.testing.functional:
    - It commits and synchronises the current transaction before and
      after the test.

    """

    def __init__(self, root_factory, wsgi_stack):
        # ZODBLayer creates DB in testSetUp method, but the middleware is
        # set up already in the `setUp` method, so we have only the
        # `root_factory` not the root itself:
        self.root_factory = root_factory
        self.wsgi_stack = wsgi_stack

    def __call__(self, environ, start_response):
        transaction.commit()
        yield from self.wsgi_stack(environ, start_response)
        self.root_factory()._p_jar.sync()


class AuthorizationMiddleware:
    """This middleware makes the WSGI application compatible with the
    HTTPCaller behavior defined in zope.app.testing.functional:
    - It modifies the HTTP Authorization header to encode user and
      password into base64 if it is Basic authentication.
    """

    def __init__(self, wsgi_stack):
        self.wsgi_stack = wsgi_stack

    def __call__(self, environ, start_response):
        # Handle authorization
        auth_key = 'HTTP_AUTHORIZATION'
        if auth_key in environ:
            environ[auth_key] = auth_header(environ[auth_key])

        # Remove unwanted headers
        def application_start_response(status, headers, exc_info=None):
            headers = list(filter(is_wanted_header, headers))
            start_response(status, headers)

        yield from self.wsgi_stack(environ, application_start_response)


class BrowserLayer(ZODBLayer):
    """This create a test layer with a test database and register a wsgi
    application to use that test database.

    You can use a WSGI version of zope.testbrowser Browser instance to access
    the application.
    """
    allowTearDown = False

    def __init__(self, package, zcml_file='ftesting.zcml',
                 name=None, features=None, allowTearDown=False):
        super().__init__(package, zcml_file, name, features)
        self.allowTearDown = allowTearDown

    def setup_middleware(self, app):
        # Override this method in subclasses of this layer in order to set up
        # WSGI middleware.
        return app

    def make_wsgi_app(self, setup_middleware=lambda a: a):
        self._application = WSGIPublisherApplication(self.db)
        return AuthorizationMiddleware(
            TransactionMiddleware(
                self.getRootFolder,
                self.setup_middleware(
                    setup_middleware(self._application)
                )
            )
        )

    def tearDown(self):
        if self.allowTearDown:
            super().tearDown()
        else:
            raise NotImplementedError


class NotInBrowserLayer(Exception):
    """The current test is not running in a layer inheriting from
    BrowserLayer.
    """


class FakeResponse:
    """This behave like a Response object returned by HTTPCaller of
    zope.app.testing.functional.
    """

    def __init__(self, response, request=None):
        self.response = response
        self.request = request

    @property
    def server_protocol(self):
        protocol = None
        if self.request is not None:
            protocol = self.request.environ.get('SERVER_PROTOCOL')
        if protocol is None:
            protocol = b'HTTP/1.0'
        if not isinstance(protocol, bytes):
            protocol = protocol.encode('latin1')
        return protocol

    def getStatus(self):
        return self.response.status_int

    def getStatusString(self):
        return self.response.status

    def getHeader(self, name, default=None):
        return self.response.headers.get(name, default)

    def getHeaders(self):
        return sorted(self.response.headerlist)

    def getBody(self):
        return self.response.body

    def __bytes__(self):
        status = self.response.status
        if not isinstance(status, bytes):
            status = status.encode('latin1')
        parts = [self.server_protocol + b' ' + status]

        headers = [(k.encode('latin1') if not isinstance(k, bytes) else k,
                    v.encode('latin1') if not isinstance(v, bytes) else v)
                   for k, v in self.getHeaders()]

        parts += [k + b': ' + v for k, v in headers]

        body = self.response.body
        if body:
            if not isinstance(body, bytes):
                body = body.encode('utf-8')
            parts += [b'', body]
        return b'\n'.join(parts)

    getOutput = __bytes__

    def __str__(self):
        return bytes(self).decode('latin-1')


def encodeMultipartFormdata(
        fields: typing.List[typing.Tuple[str, str]],
        files: typing.Optional[list] = None) -> typing.Tuple[bytes, bytes]:
    """Encode fields and files to be used in a multipart/form-data request.

    This function can be used in conjunction with `http()` (see below) to
    prepare the body of a POST request.

    Returns a tuple of content-type and content.
    """
    if files is None:
        files = []
    content_type, content = _TEST_APP_FOR_ENCODING.encode_multipart(
        fields, files)
    return content_type.encode(), content


def http(wsgi_app, string, handle_errors=True):
    request = TestRequest.from_file(BytesIO(string.lstrip()))
    request.environ['wsgi.handleErrors'] = handle_errors
    response = request.get_response(wsgi_app)
    return FakeResponse(response, request=request)


class FakeSocket:

    def __init__(self, data):
        self.data = data

    def makefile(self, mode, bufsize=None):
        return BytesIO(self.data)


class XMLRPCTestTransport(xmlrpc.client.Transport):
    """xmlrpc.client lib transport that delegates to http().

    It can be used like a normal transport, including support for basic
    authentication.
    """

    verbose = False
    handleErrors = True

    def request(self, host, handler, request_body, verbose=0):
        request = f"POST {handler} HTTP/1.0\n"
        request += "Content-Length: %i\n" % len(request_body)
        request += "Content-Type: text/xml\n"

        host, extra_headers, x509 = self.get_host_info(host)
        if extra_headers:
            request += "Authorization: {}\n".format(
                dict(extra_headers)["Authorization"])

        request += "\n" + request_body
        # XXX: http() needs to be passed a wsgi app!  where do we get a wsgi
        # app?
        response = http(request, handle_errors=self.handleErrors)

        errcode = response.getStatus()
        errmsg = response.getStatusString()
        # This is not the same way that the normal transport deals with the
        # headers.
        headers = response.getHeaders()

        if errcode != 200:
            raise xmlrpc.client.ProtocolError(
                host + handler,
                errcode, errmsg,
                headers
            )

        res = httpclient.HTTPResponse(FakeSocket(response.getBody()))
        res.begin()
        return self.parse_response(res)


def XMLRPCServerProxy(uri, transport=None, encoding=None,
                      verbose=0, allow_none=0, handleErrors=True):
    """A factory that creates a server proxy using the XMLRPCTestTransport
    by default.

    """
    if transport is None:
        transport = XMLRPCTestTransport()
    if isinstance(transport, XMLRPCTestTransport):
        transport.handleErrors = handleErrors
    return xmlrpc.client.ServerProxy(
        uri, transport, encoding, verbose, allow_none)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""zope.app.wsgi common test related classes/functions/objects.
"""
import tempfile

import zope.publisher.interfaces.browser

from zope import component
from zope import interface
from zope.app.wsgi.testlayer import BrowserLayer


@interface.implementer(zope.publisher.interfaces.browser.IBrowserPublisher)
@component.adapter(
    interface.Interface, zope.publisher.interfaces.browser.IBrowserRequest)
class FileView:

    def __init__(self, _, request):
        self.request = request

    def browserDefault(self, *_):
        return self, ()

    def __call__(self):
        self.request.response.setHeader('content-type', 'text/plain')
        fn = tempfile.mktemp()
        with open(fn, 'wb') as file:
            file.write(b"Hello\nWorld!\n")
        file = open(fn, 'rb')
        return file


class IndexView(FileView):
    def __call__(self):
        self.request.response.setHeader('content-type', 'text/html')
        return '''
            <html>
              <head>
              </head>
              <body>
                <p>This is the index</p>
              </body>
            </html>'''


class ErrorRaisingView(FileView):
    def __call__(self):
        return 1 / 0


class SillyMiddleWare:
    def __init__(self, application):
        self.application = application

    def __call__(self, environ, start_response):
        def drop_content_length_response(status, headers, exc_info=None):
            for name, value in headers:
                if name.lower() == 'content-length':
                    headers.remove((name, value))
            return start_response(status, headers, exc_info=exc_info)

        app_iter = self.application(environ, drop_content_length_response)

        # Very silly indeed:
        result = b''.join(app_iter)
        return [result.replace(
            b'<body>', b'<body><h1>Hello from the silly middleware</h1>')]


class SillyMiddleWareBrowserLayer(BrowserLayer):

    def setup_middleware(self, app):
        return SillyMiddleWare(app)
